////////////////////////////////////////////////////////////////////////////////
//
// Util.cpp
//
// The Bat! Antivirus Plugin for Softwin's Free BitDefender Lite edition
//
// Helper functions (including main scan engine)
//
//
// This is free software and source code, feel free to modify and distribute
// at your own risk.
//
// Original author: Ralf Buschmann (caranthir2@gmx.de)
//
// Version 1.0 (February 2, 2005)
//
//
//
////////////////////////////////////////////////////////////////////////////////
#include "BdLite.h"

CRITICAL_SECTION	csLogfile;
CRITICAL_SECTION	csScanner;
FILE				*Logfile = NULL;
TCHAR				*szLogfile = NULL;
extern HINSTANCE	hInstance;
extern DWORD		dwLogMax, dwLogging;
extern TCHAR		*szBdc, *szBdcDir;

////////////////////////////////////////////////////////////////////////////////
//
// InitLogFile()
//
// Initializes the log file
//
////////////////////////////////////////////////////////////////////////////////
DWORD InitLogFile(BOOL bOpen, BOOL bClear)
{
	DWORD dwError = ERROR_SUCCESS, dwSize;
	TCHAR *szDir = NULL;
	WIN32_FIND_DATA wfd;
	HANDLE hFind;
	
	if (!bOpen)
	{
		if (Logfile)
			fclose(Logfile);

		if (bClear)
			DeleteFile(szLogfile);

		if (szLogfile)
			LocalFree(szLogfile);

		DeleteCriticalSection(&csLogfile);

		Logfile = NULL;
		szLogfile = NULL;
	}
	else
	{
		if (ERROR_SUCCESS != (dwError = RegGetRegValue(
			HKEY_CURRENT_USER, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders"), TEXT("AppData"),
			NULL, (BYTE**)&szDir, &dwSize)))
		{
			return dwError;
		}

		if (szDir[lstrlen(szDir) - 1] == '\\')
			szDir[lstrlen(szDir) - 1] = 0;

		if (!FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER	|
			FORMAT_MESSAGE_FROM_STRING		|
			FORMAT_MESSAGE_ARGUMENT_ARRAY,
			TEXT("%1\\BAV_BdLite.log"),
			0,
			0,
			(LPTSTR)&szLogfile,
			0,
			&szDir
			))
		{
			LocalFree(szDir);
			return GetLastError();
		}

		LocalFree(szDir);
		
		if (INVALID_HANDLE_VALUE != (hFind = FindFirstFile(szLogfile, &wfd)))
		{
			if (dwLogMax && (wfd.nFileSizeLow >= dwLogMax))
				Logfile = _tfopen(szLogfile, TEXT("w"));
			else
				Logfile = _tfopen(szLogfile, TEXT("a+"));

			FindClose(hFind);

			if (NULL == Logfile)
				dwError = GetLastError();
		}
		else if (NULL == (Logfile = _tfopen(szLogfile, TEXT("a+"))))
		{
			return GetLastError();
		}
		
		// Protect access to log file by critical section because
		// multiple threads need access to it
		InitializeCriticalSection(&csLogfile);
	}

	return dwError;
}

////////////////////////////////////////////////////////////////////////////////
//
// LogMessage
//
// Logs a message the log file
//
////////////////////////////////////////////////////////////////////////////////
void LogMessage(TCHAR *szFormat, ...)
{
	SYSTEMTIME	SysTime;
	
	if (NULL == Logfile)
		return;

	va_list ArgList;
	
	va_start (ArgList, szFormat);

	GetLocalTime(&SysTime);

	EnterCriticalSection(&csLogfile);

	if (dwLogMax && ((DWORD)ftell(Logfile) > dwLogMax))
	{
		InitLogFile(FALSE);
		InitLogFile(TRUE);
	}

	if (NULL == Logfile)
	{
		LeaveCriticalSection(&csLogfile);
		return;
	}

	_ftprintf(Logfile, TEXT("%02d/%02d/%04d %02d:%02d:%02d: "),
			SysTime.wDay,
			SysTime.wMonth,
			SysTime.wYear,
			SysTime.wHour,
			SysTime.wMinute,
			SysTime.wSecond);

	_vftprintf(Logfile, szFormat, ArgList);

	_ftprintf(Logfile, TEXT("\n"));
	fflush(Logfile);

	LeaveCriticalSection(&csLogfile);

	va_end (ArgList);

}

////////////////////////////////////////////////////////////////////////////////
//
// Function: RegGetRegValue
//
// Fetches a value from the registry. Caller is responsible for freeing the
// returned buffer via LocalFree().
//
////////////////////////////////////////////////////////////////////////////////
DWORD RegGetRegValue(HKEY hRootKey, TCHAR *szSubKey, TCHAR *szValueName, DWORD *pdwType, BYTE **ppBuffer, DWORD *pdwSize)
{
	DWORD	dwBufferSize = 0, dwError = ERROR_SUCCESS;
	HKEY	hKey = NULL;

	__try
	{
		if (!ppBuffer || !pdwSize)
		{
			dwError = ERROR_INVALID_PARAMETER;
			__leave;
		}
		
		*ppBuffer = NULL;
		*pdwSize = 0;
		
		// If szSubKey is NULL, assume we are querying an already open key
		if (szSubKey && (ERROR_SUCCESS != (dwError = RegOpenKeyEx(
			hRootKey,
			szSubKey,
			0,
			KEY_QUERY_VALUE,
			&hKey
			))))
		{
			__leave;
		}

		if (NULL == szSubKey)
			hKey = hRootKey;

		if (ERROR_SUCCESS != (dwError = RegQueryValueEx(
			hKey,
			szValueName,
			NULL,
			pdwType,
			NULL,
			&dwBufferSize
			)))
		{
			__leave;
		}

		if (0 == dwBufferSize)
		{
			dwError = ERROR_INVALID_DATA;
			__leave;
		}

		// Required buffer size is in bytes
		if (NULL == (*ppBuffer = (BYTE*)LocalAlloc(LPTR, dwBufferSize)))
		{
			dwError = ERROR_OUTOFMEMORY;
			__leave;
		}

		ZeroMemory(*ppBuffer, dwBufferSize);

		if (ERROR_SUCCESS != (dwError = RegQueryValueEx(
			hKey,
			szValueName,
			NULL,
			pdwType,
			*ppBuffer,
			&dwBufferSize
			)))
		{
			__leave;
		}

		*pdwSize = dwBufferSize;
	}

	__finally
	{
		if (hKey && szSubKey)
			RegCloseKey(hKey);
		
		if ((ERROR_SUCCESS != dwError) && ppBuffer && *ppBuffer)
		{
			LocalFree(*ppBuffer);
			*ppBuffer = NULL;
		}
	}

	return dwError;
}

////////////////////////////////////////////////////////////////////////////////
//
// Function: RegSetRegValue
//
// Stores a value in the registry
//
////////////////////////////////////////////////////////////////////////////////
DWORD RegSetRegValue(HKEY hRootKey, TCHAR *szSubKey, TCHAR *szValueName, DWORD dwType, BYTE *pBuffer, DWORD dwSize)
{
	HKEY	hKey = NULL;
	DWORD	dwDisposition, dwError = ERROR_SUCCESS;

	__try
	{
		if (szSubKey && (ERROR_SUCCESS != (dwError = RegCreateKeyEx(
			hRootKey,
			szSubKey,
			0,
			NULL,
			0,
			KEY_SET_VALUE,
			NULL,
			&hKey,
			&dwDisposition
			))))
		{
			__leave;
		}
		
		if (NULL == szSubKey)
			hKey = hRootKey;
		
		if (ERROR_SUCCESS != (dwError = RegSetValueEx(
			hKey,
			szValueName,
			0,
			dwType,
			pBuffer,
			dwSize
			)))
		{
			__leave;
		}
	}

	__finally
	{
		if (szSubKey && hKey)
			RegCloseKey(hKey);
	}

	return dwError;
}

////////////////////////////////////////////////////////////////////////////////
//
// Function: ScanFile
//
// Scans for and optionally tries to disinfect viruses from files using
// the BitDefender command line scanner (bdc.exe).
//
////////////////////////////////////////////////////////////////////////////////
DWORD ScanFile(TCHAR *szFile, TCHAR *szVirList, BOOL bClean)
{
	HANDLE hOutputReadTmp = NULL, hOutputRead = NULL, hOutputWrite = NULL;
	HANDLE hInputWriteTmp = NULL, hInputRead = NULL, hInputWrite = NULL;
	HANDLE hErrorWrite = NULL;
	HANDLE hHeap = NULL;
	DWORD dwError = AVC_Error, dwViruses, dwCount, dwStatus;
	BOOL bFailed = FALSE, bFooter = FALSE;
	SECURITY_ATTRIBUTES sa;
	STARTUPINFO			si;
	PROCESS_INFORMATION pi;
	TCHAR szBuf[MAX_BUFFER + 1], szVirName[MAX_BUFFER + 1], szLog[MAX_PATH] = {0};
	TCHAR *szCmdLine = NULL, *p;
	TCHAR *szInserts[2] = {szFile, szLog};
	FILE *Log = NULL;

	__try
	{
		// Make sure temp file names don't conflict
		EnterCriticalSection(&csScanner);

		// Get temp filename for log
		if (!GetTempPath(MAX_PATH, szBuf))
		{
			LeaveCriticalSection(&csScanner);
			__leave;
		}
		if (!GetTempFileName(szBuf, TEXT("bdl"), 0, szLog))
		{
			LeaveCriticalSection(&csScanner);
			__leave;
		}

		LeaveCriticalSection(&csScanner);

		if (dwLogging)
		{
			if (!bClean)
				LogMessage(TEXT("Scanning %s..."), szFile);
			else
				LogMessage(TEXT("Disinfecting %s..."), szFile);
		}

		szVirList[0] = 0;

		// Set up the security attributes struct.
		sa.nLength= sizeof(SECURITY_ATTRIBUTES);
		sa.lpSecurityDescriptor = NULL;
		sa.bInheritHandle = TRUE;


		// Create the child output pipe.
		if (!CreatePipe(&hOutputReadTmp, &hOutputWrite, &sa, 0))
			__leave;

		// Create a duplicate of the output write handle for the std error
		// write handle. This is necessary in case the child application
		// closes one of its std output handles.
		if (!DuplicateHandle(
			GetCurrentProcess(), hOutputWrite, GetCurrentProcess(), &hErrorWrite,
			0, TRUE, DUPLICATE_SAME_ACCESS))
			__leave;

		// Create the child input pipe.
		if (!CreatePipe(&hInputRead, &hInputWriteTmp, &sa, 0))
			__leave;

		// Create new output read handle and the input write handles. Set
		// the Properties to FALSE. Otherwise, the child inherits the
		// properties and, as a result, non-closeable handles to the pipes
		// are created.
		if (!DuplicateHandle(
			GetCurrentProcess(), hOutputReadTmp, GetCurrentProcess(), &hOutputRead,
			0, FALSE, DUPLICATE_SAME_ACCESS))
			__leave;

		if (!DuplicateHandle(
			GetCurrentProcess(), hInputWriteTmp, GetCurrentProcess(), &hInputWrite,
			0, FALSE, DUPLICATE_SAME_ACCESS))
			__leave;

		// Close inheritable copies of the handles you do not want to be
		// inherited.
		CloseHandle(hOutputReadTmp);
		hOutputReadTmp = NULL;
		CloseHandle(hInputWriteTmp);
		hInputWriteTmp = NULL;

		ZeroMemory(&si, sizeof(si));
		ZeroMemory(&pi, sizeof(pi));

		// Set up STARTUPINFO struct to run process with redirected output
		si.cb          = sizeof(STARTUPINFO);
		si.dwFlags     = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
		si.wShowWindow = SW_HIDE;
		si.hStdOutput  = hOutputWrite;
		si.hStdInput   = hInputRead;
		si.hStdError   = hErrorWrite;

		if (!FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER	|
			FORMAT_MESSAGE_FROM_STRING		|
			FORMAT_MESSAGE_ARGUMENT_ARRAY,
			bClean ? TEXT(" \"%1\" /f /r /i \"/G=%2\" /a /W /d /alev=6") : TEXT(" \"%1\" /f /r /i \"/G=%2\" /a /W /alev=6"),
			0,
			0,
			(LPTSTR)&szCmdLine,
			0,
			szInserts
			))
			__leave;

		if (!CreateProcess(
			szBdc,
			szCmdLine,
			NULL, NULL,	TRUE, CREATE_NEW_CONSOLE, NULL,
			szBdcDir,
			&si, &pi))
			__leave;

		CloseHandle(pi.hThread);
		CloseHandle(hOutputWrite);
		hOutputWrite = NULL;
		CloseHandle(hInputRead);
		hInputRead = NULL;
		CloseHandle(hErrorWrite);
		hErrorWrite = NULL;

		dwViruses = 0;
		bFailed = FALSE;
		bFooter = FALSE;
		dwCount = COUNT_MAX;
		dwError = AVC_OK;
		
		WaitForSingleObject(pi.hProcess, INFINITE);
		GetExitCodeProcess(pi.hProcess, &dwStatus);
		CloseHandle(pi.hProcess);

		if (NULL == (Log = _tfopen(szLog, TEXT("rt"))))
			__leave;

		// Parse log file
		while (_fgetts(szBuf, MAX_BUFFER, Log) && (0 != szBuf[0]))
		{
			// Skip header
			if (_tcsstr(szBuf, TEXT("//")) == szBuf)
				continue;

			// Skip footer
			if (_tcsstr(szBuf, TEXT("Results:")) == szBuf)
				break;

			// Remove trailing line breaks and blanks
			p = szBuf + lstrlen(szBuf) - 1;
			while ((p > szBuf) && ((*p == '\r') || (*p == '\n') || (*p == ' ')))
			{
				*p = 0;
				p--;
			}

			// Result of disinfection
			if (bClean && (dwCount == 0))
			{
				dwCount = COUNT_MAX;

				p = _tcsstr(szBuf, TEXT("\t"));

				if ((p == NULL) || ((_tcsstr(p, DISINFECTED_KEY) == NULL) && (_tcsstr(p, DELETED_KEY) == NULL)))
				{
					if (!bFailed)
						szVirList[0] = 0;
					bFailed = TRUE;

					if ((lstrlen(szVirName) + lstrlen(szVirList) + 2) < MAX_BUFFER)
					{
						lstrcat(szVirList, szVirName);
						lstrcat(szVirList, TEXT("\n"));
					}
				}
				else if (((lstrlen(szVirName) + lstrlen(szVirList) + 2) < MAX_BUFFER) && !bFailed)
				{
					lstrcat(szVirList, szVirName);
					lstrcat(szVirList, TEXT("\n"));
				}
				
				if (dwLogging)
					LogMessage(TEXT("\t%s"), szBuf);
			}
			
			// Look for the "infected:" or "suspected:" keywords
			p = NULL;
			if ((p = _tcsstr(szBuf, INFECTED_KEY)) || (p = _tcsstr(szBuf, SUSPECTED_KEY)))
			{
				dwViruses++;

				// Point at virus name
				p = _tcsstr(p, TEXT(":"));
				while (*p)
				{
					p++;
					if (*p != ' ')
						break;
				}
				
				// Log virus name
				if (!bClean && ((lstrlen(p) + lstrlen(szVirList) + 2) < MAX_BUFFER))
				{
					lstrcat(szVirList, p);
					lstrcat(szVirList, TEXT("\n"));

				}
				else
					lstrcpyn(szVirName, p, MAX_BUFFER);

				dwCount = 0;
				
				if (_tcsstr(szBuf, SUSPECTED_KEY) && (dwError != AVC_VirusFound))
					dwError = AVC_Suspicious;
				else
					dwError = AVC_VirusFound;
				
				if (dwLogging)
					LogMessage(TEXT("\t%s"), szBuf);
			}
		}

		if (bClean)
		{
			// AVC_CantBeCured causes TB! to delete the attachment, so that's what we
			// return if the scanner deleted the file. Otherwise AVC_VirusFound causes
			// TB! to use the modified attachment. Returning AVC_OK stores the original
			// (infected!) attachment.
			if (bFailed || (INVALID_FILE_ATTRIBUTES == GetFileAttributes(szFile)))
				dwError = AVC_CantBeCured;
			else
				dwError = AVC_VirusFound;
		}

		// Sanity check
		if ((dwViruses == 0) && (dwStatus != 0))
		{
			lstrcpy(&szVirList[0], TEXT("Internal error"));
			if (dwLogging)
				LogMessage(TEXT("\tInternal error!"));
			dwError = AVC_Error;
		}
	}

	__finally {

		if (hOutputReadTmp)
			CloseHandle(hOutputReadTmp);
		if (hInputWriteTmp)
			CloseHandle(hInputWriteTmp);
		if (hOutputRead)
			CloseHandle(hOutputRead);
		if (hInputRead)
			CloseHandle(hInputRead);
		if (hOutputWrite)
			CloseHandle(hOutputWrite);
		if (hInputWrite)
			CloseHandle(hInputWrite);
		if (hErrorWrite)
			CloseHandle(hErrorWrite);
		if (szCmdLine)
			LocalFree(szCmdLine);
		if (Log)
		{
			fclose(Log);
			DeleteFile(szLog);
		}
	}

	return dwError;
}
